
MatePanelAppletFrameDBus > MatePanelAppletFrameDBus {
	border-style: none;
	background-repeat: no-repeat;
	background-position: left;
	color: @theme_fg_color;
	background-image: -gtk-recolor(url("panel-grid-symbolic.svg"));
	background-size: 12px 22px;
}

.mate-panel-menu-icon-box{
	min-width: 16px;
}
